#define STRICT
#define WIN32_LEAN_AND_MEAN

#define TERRAIN_VERTEXCOUNT (TERRAINX * TERRAINZ)
#define TERRAINX 4
#define TERRAINZ 4
#define TERRAIN_TRIANGLECOUNT ((TERRAINX-1)*(TERRAINZ-1)*2)
#define TERRAINSIZE 25.0f
#define TERRAINHEIGHT 10


#include "tokamak.h"
#pragma comment(lib, "tokamak.lib")
#include <GL/glut.h>
#include <windows.h>
#include "matrix4x4f.h"
#include "vector3f.h"

// This macro converts a Tokamak matrix (neT3) to my custom matrix type (matrix4x4f).
#define NET3_TO_MATRIX4X4f( tmat )                                        \
matrix4x4f( tmat.rot[0][0], tmat.rot[1][0], tmat.rot[2][0], tmat.pos[0],  \
            tmat.rot[0][1], tmat.rot[1][1], tmat.rot[2][1], tmat.pos[1],  \
            tmat.rot[0][2], tmat.rot[1][2], tmat.rot[2][2], tmat.pos[2],  \
            0.0f,          0.0f,          0.0f,          1.0f           );


extern "C"	void initPhysics(float v0, float v1, float v2);
extern "C"	void updatePhysics(float v0, float v1, float v2);
extern "C"	void TokamakRender(float v0, float v1, float v2);
extern "C"	void Tokamak_TriMesh (float pos1, float pos2, float pos3);

extern "C"  float GetOriginX(float ass, int rt0);
extern "C"  float GetOriginY(float ass1, int rt1);
extern "C"  float GetOriginZ(float ass2, int rt2);

extern "C"  float GetVelocityX(float ass, int ent0);
extern "C"  float GetVelocityY(float ass1, int ent1);
extern "C"  float GetVelocityZ(float ass2, int ent2);

extern "C"  float GetPlayerVelocityX (float pvel0);
extern "C"  float GetPlayerVelocityY (float pvel1);
extern "C"  float GetPlayerVelocityZ (float pvel2);

extern "C"  float GetPlayerOriginX(float ass);
extern "C"  float GetPlayerOriginY(float ass1);
extern "C"  float GetPlayerOriginZ(float ass2);

HWND  g_hWnd = NULL;
HDC	  g_hDC  = NULL;
HGLRC g_hRC  = NULL;


// Tokamak related globals ----------------------------------------------------
LARGE_INTEGER g_performanceFrequency; // Used to store the performance frequency of the target box
LARGE_INTEGER g_startTime;            // Used to store the tick count at the beginning of each game loop
LARGE_INTEGER g_endTime;              // Used to store the tick count at the end of each game loop
float		  g_fFrameTime;           // Used to store the resulting time taken from start to end of each game loop (1 frame)
float		  g_fTimeSinceLastUpdate = 0.0f;
const float	  CONST_FRAME_INTERVAL = 0.0166666666f; //0.01666666f;

const int NUM_RED_CUBES = 69;
const int NUM_BLUE_PLATFORMS = 1;

neSimulator    *g_simulator = NULL;
neRigidBody    *g_redCubes[NUM_RED_CUBES];
neAnimatedBody *g_bluePlatform;
neAnimatedBody *gFloor = NULL;
float g_fTiltOfBluePlatform = 0.0f;

//-----------------------------------------------------------------------------

struct Vertex
{
	// GL_N3F_V3F
	float nx, ny, nz;
	float x, y, z;
};

Vertex g_cubeVertices[] =
{
	{ 0.0f, 0.0f, 10.0f,  -9.5f,-9.5f, 9.5f },
	{ 0.0f, 0.0f, 10.0f,   9.5f,-9.5f, 9.5f },
	{ 0.0f, 0.0f, 10.0f,   9.5f, 9.5f, 9.5f },
	{ 0.0f, 0.0f, 10.0f,  -9.5f, 9.5f, 9.5f },

	{ 0.0f, 0.0f,-10.0f,  -9.5f,-9.5f,-9.5f },
	{ 0.0f, 0.0f,-10.0f,  -9.5f, 9.5f,-9.5f },
	{ 0.0f, 0.0f,-10.0f,   9.5f, 9.5f,-9.5f },
	{ 0.0f, 0.0f,-10.0f,   9.5f,-9.5f,-9.5f },

	{ 0.0f, 10.0f, 0.0f,  -9.5f, 9.5f,-9.5f },
	{ 0.0f, 10.0f, 0.0f,  -9.5f, 9.5f, 9.5f },
	{ 0.0f, 10.0f, 0.0f,   9.5f, 9.5f, 9.5f },
	{ 0.0f, 10.0f, 0.0f,   9.5f, 9.5f,-9.5f },

	{ 0.0f,-10.0f, 0.0f,  -9.5f,-9.5f,-9.5f },
	{ 0.0f,-10.0f, 0.0f,   9.5f,-9.5f,-9.5f },
	{ 0.0f,-10.0f, 0.0f,   9.5f,-9.5f, 9.5f },
	{ 0.0f,-10.0f, 0.0f,  -9.5f,-9.5f, 9.5f },

	{ 10.0f, 0.0f, 0.0f,   9.5f,-9.5f,-9.5f },
	{ 10.0f, 0.0f, 0.0f,   9.5f, 9.5f,-9.5f },
	{ 10.0f, 0.0f, 0.0f,   9.5f, 9.5f, 9.5f },
	{ 10.0f, 0.0f, 0.0f,   9.5f,-9.5f, 9.5f },

	{-10.0f, 0.0f, 0.0f,  -9.5f,-9.5f,-9.5f },
	{-10.0f, 0.0f, 0.0f,  -9.5f,-9.5f, 9.5f },
	{-10.0f, 0.0f, 0.0f,  -9.5f, 9.5f, 9.5f },
	{-10.0f, 0.0f, 0.0f,  -9.5f, 9.5f,-9.5f } 
};


//-----------------------------------------------------------------------------
// Name: initPhysics()
// Desc: Real stuff starts here:
//-----------------------------------------------------------------------------

void initPhysics(float v0, float v1, float v2)  //v0,v1,v2 not used anymore
{
    QueryPerformanceFrequency( &g_performanceFrequency );
    QueryPerformanceCounter( &g_startTime );

	//
	// Setup the physics environment...
	//

	neV3 gravity;
	gravity.Set( 0.0f, 0.0f, -100.0f );  //was (0,-10,0) but quake is tilted on a diff axis

	neSimulatorSizeInfo sizeInfo;

	sizeInfo.rigidBodiesCount        = NUM_RED_CUBES + 1;  //add one for player (more for all models later)
	sizeInfo.animatedBodiesCount     = NUM_BLUE_PLATFORMS + 1; //add 1 for the bsp mesh
	sizeInfo.rigidParticleCount      = 0;
	sizeInfo.controllersCount        = 0;
	// Use the following formula to compute the number of overlapped-pairs required: 
	//(num_rigid_bodies * num_animated_bodies) + num_rigid_bodies * (num_rigid_bodies - 1) / 2
	sizeInfo.overlappedPairsCount    = ((NUM_RED_CUBES +1) * NUM_BLUE_PLATFORMS) + (NUM_RED_CUBES + 1) * ((NUM_RED_CUBES + 1) - 1) / 2;
	sizeInfo.geometriesCount         = (NUM_RED_CUBES + 1) + NUM_BLUE_PLATFORMS;
	sizeInfo.constraintsCount        = 0;
	sizeInfo.constraintSetsCount     = 0;
	sizeInfo.constraintBufferSize    = 0;
	sizeInfo.sensorsCount            = 0;
	sizeInfo.terrainNodesStartCount  = 0;
	sizeInfo.terrainNodesGrowByCount = 0;

	g_simulator = neSimulator::CreateSimulator( sizeInfo, NULL, &gravity );

	for( int i = 0; i < NUM_RED_CUBES + 1; ++i )
	{
		g_redCubes[i] = g_simulator->CreateRigidBody();
		neGeometry *cubeGeometry = g_redCubes[i]->AddGeometry();

		neV3 cubeSize;
		cubeSize.Set( 19.0f, 19.0f, 19.0f ); // Unit size  (physics BBOX, not model dimensions)
		cubeGeometry->SetBoxSize( cubeSize );
		//need to set the size based on QuakeC bounding box

		g_redCubes[i]->UpdateBoundingInfo();

		float fMass = 100.1f;
		g_redCubes[i]->SetInertiaTensor( neBoxInertiaTensor( cubeSize, fMass ) );
		g_redCubes[i]->SetMass( fMass );

		neV3 position;
		//cube positioning:
		if( i == 0 )
		{
			position.Set( 0.0f, 0.0f, 39.5f );  //first cube
		}
		else if((i != 0) && (i < NUM_RED_CUBES + 1) && (i !=5))
			position.Set( 0.0f, 0.0f, (39.5f * (float)(i + 1)) );  //rest of cubes
		else if (i == 5) //(i >= NUM_RED_CUBES)
		{			
			cubeSize.Set( 19.0f, 19.0f, 19.0f ); // Unit size  (physics BBOX, not model dimensions)
			float qo0, qo1, qo2;
			qo0 = GetOriginX(0.0, 0);
			qo1 = GetOriginY(0.0, 0);
			qo2 = GetOriginZ(0.0, 0);
			//position.Set(qo0, qo1, qo2); //q1 position, getting world origin, not player's??
			position.Set(v0, v1, v2);
		}
		// Uncomment this to stack the cubes using their size instead of 
		// dropping them into place.
		//position.Set( 0.0f, (cubeSize[1] * (i + 1)), 0.0f );

		g_redCubes[i]->SetPos( position );
	}

	//
	// Physics for the blue platform...  //switch this crap to bsp mesh
	//

	g_bluePlatform = g_simulator->CreateAnimatedBody();

	neGeometry *platformGeometry = g_bluePlatform->AddGeometry();

	platformGeometry->SetBoxSize( 2000.0f, 2000.0f, 3.5f );

	g_bluePlatform->UpdateBoundingInfo();

	neV3 position;
	position.Set( 0.0f, 0.0f, -45.0f );

	g_bluePlatform->SetPos( position );	
}


//-----------------------------------------------------------------------------
// Name: updatePhysics()
// Desc: 
//-----------------------------------------------------------------------------
void updatePhysics(float v0, float v1, float v2)
{
        //
    // Before we advance the simulation, lets update the orientation of our  
    // blue platform just in case the user adjusted it.
    //

    neT3 rotMatrix;
    neV3 rotation;
    rotation.Set( g_fTiltOfBluePlatform, 0.0f, 0.0f );
    rotMatrix.rot.RotateXYZ( rotation );
    g_bluePlatform->SetRotation( rotMatrix.rot );


//loop below for each entity:
	for (int x=0; x < NUM_RED_CUBES; ++x)
	{

	neM3 rot;
	neV3 r; r.Set(0,0,0);
	rot.RotateXYZ(r);

	float qo0, qo1, qo2;
	float qv0, qv1, qv2;
	float final0, final1, final2;

	qo0 = GetOriginX(0.0, x);
	qo1 = GetOriginY(0.0, x);
	qo2 = GetOriginZ(0.0, x);

	qv0 = GetVelocityX(0.0, x);
	qv1 = GetVelocityY(0.0, x);
	qv2 = GetVelocityZ(0.0, x);

	neV3 position, CurrentOrigin;
	neV3 vel;

	//FIXME: if quake entity is non-active, set quake vel = tokamak movement (create SetVelocity function)

	CurrentOrigin = g_redCubes[x]->GetPos();
	if (abs(qo0 - CurrentOrigin[0]) > 10) final0 = qo0;
	if (abs(qo1 - CurrentOrigin[1]) > 10) final1 = qo1;
	if (abs(qo2 - CurrentOrigin[2]) > 10) final2 = qo2;
	if ((abs(qo0 - CurrentOrigin[0]) > 10) || (abs(qo0 - CurrentOrigin[0]) > 10) || (abs(qo0 - CurrentOrigin[0]) > 10))
	{
		position.Set(qo0, qo1, qo2);
		g_redCubes[x]->SetPos(position);
		vel.Set (0,0,0); 	
	}	
//	vel.Set (final0, final1, final2 + 0.25);
	vel.Set (qv0, qv1, qv2 + 0.25); //add .25 per (10) of gravity		
	g_redCubes[x]->SetVelocity(vel);
	g_redCubes[x]->SetRotation(rot);

	}
// ^end entity looping above^


	
	float qo0, qo1, qo2;
	float final0, final1, final2;
	float pvel0, pvel1, pvel2;

	pvel0 = GetPlayerVelocityX(0.0);
	pvel1 = GetPlayerVelocityY(0.0);
	pvel2 = GetPlayerVelocityZ(0.0);

	qo0 = GetPlayerOriginX(0.0);
	qo1 = GetPlayerOriginY(0.0);
	qo2 = GetPlayerOriginZ(0.0);

	neV3 position, CurrentOrigin;
	neV3 vel;


	CurrentOrigin = g_redCubes[5]->GetPos();
	if (abs(qo0 - CurrentOrigin[0]) > 10) final0 = qo0;
	if (abs(qo1 - CurrentOrigin[1]) > 10) final1 = qo1;
	if (abs(qo2 - CurrentOrigin[2]) > 10) final2 = qo2;
	if ((abs(qo0 - CurrentOrigin[0]) > 10) || (abs(qo0 - CurrentOrigin[0]) > 10) || (abs(qo0 - CurrentOrigin[0]) > 10))
	{
		position.Set(qo0, qo1, qo2);
		g_redCubes[5]->SetPos(position);
		vel.Set (0,0,0); 	
	}
//	vel.Set (0,0,0); 	
//	vel.Set (final0, final1, final2);
	vel.Set (pvel0, pvel1, pvel2 + 0.25); //add .25 per (10) of gravity
	//if (qv2 == 0) vel.Set (0,0,0); 	
	g_redCubes[5]->SetVelocity(vel);

	



    QueryPerformanceCounter( &g_endTime );

    // Calculate the elapsed time since the physics was last updated.
    g_fFrameTime = ( (float)g_endTime.QuadPart - (float)g_startTime.QuadPart ) 
                   / g_performanceFrequency.QuadPart;

    g_fTimeSinceLastUpdate += g_fFrameTime;

	// If the time is greater than 1/60th of a second update the physics
    if( g_fTimeSinceLastUpdate > CONST_FRAME_INTERVAL )
    {
        g_simulator->Advance( CONST_FRAME_INTERVAL, 1, NULL );
        g_fTimeSinceLastUpdate -= CONST_FRAME_INTERVAL;
    }

    QueryPerformanceCounter( &g_startTime );
}



void Tokamak_TriMesh (float pos1, float pos2, float pos3)
{
	int iTriIndex;
    float fTerrainX, fTerrainZ;
    int x,z;

    neTriangleMesh triMesh;
    neV3 triVertices[TERRAIN_VERTEXCOUNT];
    neTriangle triData[TERRAIN_TRIANGLECOUNT];   
    triMesh.vertexCount = TERRAIN_VERTEXCOUNT;
	
/* loop to add vertex values to array for terrain map (not bsp map):
	fTerrainZ = -TERRAINSIZE/2;
    for (z=0; z<TERRAINZ; z++)
    {
        fTerrainX = -TERRAINSIZE/2;
        for (x=0; x<TERRAINX; x++)
        {
            triVertices[x + z*TERRAINX].Set(fTerrainX, gfTerrainHeights[x][z], fTerrainZ);
            fTerrainX += (TERRAINSIZE / (TERRAINX-1));
        }
        fTerrainZ += (TERRAINSIZE / (TERRAINZ-1));
    }
    triMesh.vertices = triVertices;
	triMesh.triangleCount = TERRAIN_TRIANGLECOUNT;
*/


// Set the triangle values
    iTriIndex = 0;
    for (x=0; x<TERRAINX-1; x++)
    {
        for (z=0; z<TERRAINZ-1; z++)
        {
            // Make the triangle for the first half of the quad
            triData[iTriIndex].indices[0] = x     + z    *TERRAINX;
            triData[iTriIndex].indices[1] = x     + (z+1)*TERRAINX;
            triData[iTriIndex].indices[2] = (x+1) + z    *TERRAINX;
            triData[iTriIndex].materialID = 0;
            triData[iTriIndex].flag = neTriangle::NE_TRI_TRIANGLE;
            // Move to the next triangle in the array
            iTriIndex += 1;
            // Make the triangle for the second half of the quad
            triData[iTriIndex].indices[0] = x     + (z+1)*TERRAINX;
            triData[iTriIndex].indices[1] = (x+1) + (z+1)*TERRAINX;
            triData[iTriIndex].indices[2] = (x+1) + z    *TERRAINX;
            triData[iTriIndex].materialID = 0;
            triData[iTriIndex].flag = neTriangle::NE_TRI_TRIANGLE;
            // Move to the next triangle in the array
            iTriIndex += 1;
        }
    }
    
    triMesh.triangles = triData;
	//g_simulator->SetTerrainMesh(&triMesh);  //crashes right now, no vertex array loaded

}

//-----------------------------------------------------------------------------
// Name: TokamakRender()
// Desc: 
//-----------------------------------------------------------------------------
void TokamakRender(float v0, float v1, float v2)
{
    updatePhysics(v0, v1, v2);

//    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

//	getRealTimeUserInput();

  //  glMatrixMode( GL_MODELVIEW );
    //glLoadIdentity();   //loads viewmodel
	//updateViewMatrix();

	//GLfloat position_light0[] = { -0.5f, 0.8f, 0.7f, 0.0f };
	//glLightfv( GL_LIGHT0, GL_POSITION, position_light0 );

    //
	// Render our chain, which is composed of 30 red cubes joined by ball-joints
	//

	matrix4x4f matWorld;
	matrix4x4f matScale;

	for( int i = 0; i < NUM_RED_CUBES + 1; i++ )
	{
		if(g_redCubes[i])
		{
            matWorld = NET3_TO_MATRIX4X4f( g_redCubes[i]->GetTransform() );			
            glPushMatrix();
            {			
                glMatrixMode( GL_MODELVIEW );
                glMultMatrixf( matWorld.m );
				glColor3f( 1.0f, 0.0f, 1.0f );
                glInterleavedArrays( GL_N3F_V3F, 0, g_cubeVertices );                
				glDrawArrays( GL_QUADS, 0, 24 );				
            }
            glPopMatrix();
		}
	}


	//
    // Render the large blue platform, which will keep the chain from falling
    // - this crap isn't needed
/*
    matWorld = NET3_TO_MATRIX4X4f( g_bluePlatform->GetTransform() );

    matScale.scale( vector3f( 10.0f, 10.0f, 2.0f ) ); // Scale it so it matches the Tokamak object
    matWorld = matWorld * matScale;

    glPushMatrix();
    {
        glMatrixMode( GL_MODELVIEW );
        glMultMatrixf( matWorld.m );
		glColor3f( 0.0f, 0.0f, 1.0f );
        glInterleavedArrays( GL_N3F_V3F, 0, g_cubeVertices );
        glDrawArrays( GL_QUADS, 0, 24 );
    }
    glPopMatrix(); */
 	//
	// That's all... just swap the buffers...
	//

//	SwapBuffers( g_hDC );
}